@@ -18,33 +18,30 @@
  */
 package org.apache.felix.karaf.main;
 
-import java.io.File;
-import java.io.FileNotFoundException;
-import java.io.IOException;
-import java.io.InputStream;
-import java.io.InputStreamReader;
-import java.io.BufferedReader;
+import java.io.*;
+import java.lang.management.ManagementFactory;
+import java.lang.management.RuntimeMXBean;
+import java.net.InetAddress;
 import java.net.MalformedURLException;
+import java.net.ServerSocket;
+import java.net.Socket;
 import java.net.URL;
 import java.net.URLClassLoader;
+import java.security.AccessControlException;
 import java.security.Provider;
 import java.security.Security;
-import java.util.ArrayList;
-import java.util.Enumeration;
-import java.util.HashMap;
-import java.util.Iterator;
-import java.util.Map;
-import java.util.Properties;
-import java.util.StringTokenizer;
+import java.util.*;
+import java.util.logging.Level;
 import java.util.logging.Logger;
-import java.util.concurrent.CountDownLatch;
 import java.lang.reflect.Method;
-import java.lang.reflect.Constructor;
+import java.util.regex.Matcher;
+import java.util.regex.Pattern;
 
 import org.apache.felix.karaf.main.Utils;
 import org.osgi.framework.Bundle;
 import org.osgi.framework.BundleContext;
 import org.osgi.framework.Constants;
+import org.osgi.framework.FrameworkEvent;
 import org.osgi.framework.ServiceReference;
 import org.osgi.framework.launch.Framework;
 import org.osgi.framework.launch.FrameworkFactory;
@@ -113,11 +110,6 @@ public class Main {
      */
     public static final String PROPERTY_CONVERT_TO_MAVEN_URL = "karaf.maven.convert";
 
-    /**
-     * ConfigAdmin properties directory
-     */
-    public static final String PROPERTY_CM_DIRECTORY = "karaf.cm.dir";
-
     /**
      * If a lock should be used before starting the runtime
      */
@@ -132,6 +124,22 @@ public class Main {
 
     public static final String PROPERTY_LOCK_LEVEL = "karaf.lock.level";
 
+    public static final String DEFAULT_REPO = "karaf.default.repository";
+    
+    public static final String KARAF_FRAMEWORK = "karaf.framework";
+
+    public static final String KARAF_SHUTDOWN_PORT = "karaf.shutdown.port";
+
+    public static final String KARAF_SHUTDOWN_HOST = "karaf.shutdown.host";
+
+    public static final String KARAF_SHUTDOWN_PORT_FILE = "karaf.shutdown.port.file";
+
+    public static final String KARAF_SHUTDOWN_COMMAND = "karaf.shutdown.command";
+
+    public static final String KARAF_SHUTDOWN_PID_FILE = "karaf.shutdown.pid.file";
+
+    public static final String DEFAULT_SHUTDOWN_COMMAND = "SHUTDOWN";
+
     public static final String PROPERTY_LOCK_CLASS_DEFAULT = SimpleFileLock.class.getName();
 
     Logger LOG = Logger.getLogger(this.getClass().getName());
@@ -164,56 +172,66 @@ public class Main {
         System.setProperty(PROP_KARAF_BASE, karafBase.getPath());
 
         // Load system properties.
-        loadSystemProperties();
+        loadSystemProperties(karafBase);
+
+        updateInstancePid();
 
         // Read configuration properties.
         configProps = loadConfigProperties();
         BootstrapLogManager.setProperties(configProps);
         LOG.addHandler(BootstrapLogManager.getDefaultHandler());
-
+        
         // Copy framework properties from the system properties.
         Main.copySystemProperties(configProps);
 
+        updateClassLoader(configProps);
+
         processSecurityProperties(configProps);
 
         if (configProps.getProperty(Constants.FRAMEWORK_STORAGE) == null) {
             File storage = new File(karafBase.getPath(), "data/cache");
-            storage.mkdirs();
+            try {
+                storage.mkdirs();
+            } catch (SecurityException se) {
+                throw new Exception(se.getMessage()); 
+            }
             configProps.setProperty(Constants.FRAMEWORK_STORAGE, storage.getAbsolutePath());
         }
+        
+        defaultStartLevel = Integer.parseInt(configProps.getProperty(Constants.FRAMEWORK_BEGINNING_STARTLEVEL));
+        lockStartLevel = Integer.parseInt(configProps.getProperty(PROPERTY_LOCK_LEVEL, Integer.toString(lockStartLevel)));
+        lockDelay = Integer.parseInt(configProps.getProperty(PROPERTY_LOCK_DELAY, Integer.toString(lockDelay)));
+        configProps.setProperty(Constants.FRAMEWORK_BEGINNING_STARTLEVEL, Integer.toString(lockStartLevel));
+        // Start up the OSGI framework
 
-        try {
-            defaultStartLevel = Integer.parseInt(configProps.getProperty(Constants.FRAMEWORK_BEGINNING_STARTLEVEL));
-            lockStartLevel = Integer.parseInt(configProps.getProperty(PROPERTY_LOCK_LEVEL, Integer.toString(lockStartLevel)));
-            lockDelay = Integer.parseInt(configProps.getProperty(PROPERTY_LOCK_DELAY, Integer.toString(lockDelay)));
-            configProps.setProperty(Constants.FRAMEWORK_BEGINNING_STARTLEVEL, Integer.toString(lockStartLevel));
-            // Start up the OSGI framework
-
-            InputStream is = getClass().getResourceAsStream("/META-INF/services/" + FrameworkFactory.class.getName());
-            BufferedReader br = new BufferedReader(new InputStreamReader(is, "UTF-8"));
-            String factoryClass = br.readLine();
-            br.close();
-            FrameworkFactory factory = (FrameworkFactory) getClass().getClassLoader().loadClass(factoryClass).newInstance();
-            framework = factory.newFramework(new StringMap(configProps, false));
-            framework.start();
-            processAutoProperties(framework.getBundleContext());
-            // Start lock monitor
-            new Thread() {
-                public void run() {
-                    lock(configProps);
-                }
-            }.start();
-        }
-        catch (Exception ex) {
-            setExitCode(-1);
-            throw new Exception("Could not create framework", ex);
-        }
+        InputStream is = getClass().getResourceAsStream("/META-INF/services/" + FrameworkFactory.class.getName());
+        BufferedReader br = new BufferedReader(new InputStreamReader(is, "UTF-8"));
+        String factoryClass = br.readLine();
+        br.close();
+        FrameworkFactory factory = (FrameworkFactory) getClass().getClassLoader().loadClass(factoryClass).newInstance();
+        framework = factory.newFramework(new StringMap(configProps, false));
+        framework.start();
+        processAutoProperties(framework.getBundleContext());
+        // Start lock monitor
+        new Thread() {
+            public void run() {
+                lock(configProps);
+            }
+        }.start();
     }
 
     public void destroy(boolean await) throws Exception {
+        if (framework == null) {
+            return;
+        }
         try {
             if (await) {
-                framework.waitForStop(0);
+                while (true) {
+                    FrameworkEvent event = framework.waitForStop(0);
+                    if (event.getType() != FrameworkEvent.STOPPED_UPDATE) {
+                        break;
+                    }
+                }
             }
             exiting = true;
             if (framework.getState() == Bundle.ACTIVE) {
@@ -305,9 +323,15 @@ public class Main {
         final Main main = new Main(args);
         try {
             main.launch();
+        } catch (Throwable ex) {
+            main.setExitCode(-1);
+            System.err.println("Could not create framework: " + ex);
+            ex.printStackTrace();
+        }        
+        try {
             main.destroy(true);
-        }
-        catch (Throwable ex) {
+        } catch (Throwable ex) {
+            main.setExitCode(-2);
             System.err.println("Error occured shutting down framework: " + ex);
             ex.printStackTrace();
         } finally {
@@ -329,6 +353,67 @@ public class Main {
         }
     }
 
+    private void updateInstancePid() {
+        try {
+            String instanceName = System.getProperty("karaf.name");
+            String pid = ManagementFactory.getRuntimeMXBean().getName();
+            if (pid.indexOf('@') > 0) {
+                pid = pid.substring(0, pid.indexOf('@'));
+            }
+            
+            boolean isRoot = karafHome.equals(karafBase);
+            
+            if (instanceName != null) {
+                String storage = System.getProperty("storage.location");
+                if (storage == null) {
+                    throw new Exception("System property 'storage.location' is not set. \n" +
+                        "This property needs to be set to the full path of the instance.properties file.");
+                }
+                File storageFile = new File(storage);
+                File propertiesFile = new File(storageFile, "instance.properties");
+                Properties props = new Properties();
+                if (propertiesFile.exists()) {
+                    FileInputStream fis = new FileInputStream(propertiesFile);
+                    props.load(fis);
+                    int count = Integer.parseInt(props.getProperty("count"));
+                    for (int i = 0; i < count; i++) {
+                        String name = props.getProperty("item." + i + ".name");
+                        if (name.equals(instanceName)) {
+                            props.setProperty("item." + i + ".pid", pid);
+                            FileOutputStream fos = new FileOutputStream(propertiesFile);
+                            props.store(fos, null);
+                            fis.close();
+                            fos.close();
+                            return;
+                        }
+                    }
+                    fis.close();
+                    if (!isRoot) {
+                        throw new Exception("Instance " + instanceName + " not found");
+                    } 
+                } else if (isRoot) {
+                    if (!propertiesFile.getParentFile().exists()) {
+                        try {
+                            propertiesFile.getParentFile().mkdirs();
+                        } catch (SecurityException se) {
+                            throw new Exception(se.getMessage());
+                        }
+                    }
+                    props.setProperty("count", "1");
+                    props.setProperty("item.0.name", instanceName);
+                    props.setProperty("item.0.loc", karafHome.getAbsolutePath());
+                    props.setProperty("item.0.pid", pid);
+                    props.setProperty("item.0.root", "true");
+                    FileOutputStream fos = new FileOutputStream(propertiesFile);
+                    props.store(fos, null);
+                    fos.close();
+                }
+            }
+        } catch (Exception e) {
+            System.err.println("Unable to update instance pid: " + e.getMessage());
+        }
+    }
+
     /**
      * <p/>
      * Processes the auto-install and auto-start properties from the
@@ -359,46 +444,7 @@ public class Main {
         // the start level to which the bundles are assigned is specified by
         // appending a ".n" to the auto-install property name, where "n" is
         // the desired start level for the list of bundles.
-        for (Iterator i = configProps.keySet().iterator(); i.hasNext();) {
-            String key = (String) i.next();
-
-            // Ignore all keys that are not the auto-install property.
-            if (!key.startsWith(PROPERTY_AUTO_INSTALL)) {
-                continue;
-            }
-
-            // If the auto-install property does not have a start level,
-            // then assume it is the default bundle start level, otherwise
-            // parse the specified start level.
-            int startLevel = sl.getInitialBundleStartLevel();
-            if (!key.equals(PROPERTY_AUTO_INSTALL)) {
-                try {
-                    startLevel = Integer.parseInt(key.substring(key.lastIndexOf('.') + 1));
-                }
-                catch (NumberFormatException ex) {
-                    System.err.println("Invalid property: " + key);
-                }
-            }
-
-            StringTokenizer st = new StringTokenizer(configProps.getProperty(key), "\" ", true);
-            if (st.countTokens() > 0) {
-                String location = null;
-                do {
-                    location = nextLocation(st);
-                    if (location != null) {
-                        try {
-                            String[] parts = convertToMavenUrlsIfNeeded(location, convertToMavenUrls);
-                            Bundle b = context.installBundle(parts[0], new URL(parts[1]).openStream());
-                            sl.setBundleStartLevel(b, startLevel);
-                        }
-                        catch (Exception ex) {
-                            System.err.println("Auto-properties install: " + ex);
-                        }
-                    }
-                }
-                while (location != null);
-            }
-        }
+        autoInstall(PROPERTY_AUTO_INSTALL, context, sl, convertToMavenUrls);
 
         // The auto-start property specifies a space-delimited list of
         // bundle URLs to be automatically installed and started into each
@@ -407,19 +453,32 @@ public class Main {
         // where "n" is the desired start level for the list of bundles.
         // The following code starts bundles in two passes, first it installs
         // them, then it starts them.
+            List<Bundle> bundlesToStart = autoInstall(PROPERTY_AUTO_START, context, sl, convertToMavenUrls);
+        // Now loop through and start the installed bundles.
+        for (Bundle b : bundlesToStart) {
+            try {
+                b.start();
+            }
+            catch (Exception ex) {
+                System.err.println("Auto-properties start: " + ex);
+            }
+        }
+    }
+
+    private List<Bundle> autoInstall(String propertyPrefix, BundleContext context, StartLevel sl, boolean convertToMavenUrls) {
+        Map<Integer, String> autoStart = new TreeMap<Integer, String>();
+        List<Bundle> bundles = new ArrayList<Bundle>();
         for (Iterator i = configProps.keySet().iterator(); i.hasNext();) {
             String key = (String) i.next();
-
             // Ignore all keys that are not the auto-start property.
-            if (!key.startsWith(PROPERTY_AUTO_START)) {
+            if (!key.startsWith(propertyPrefix)) {
                 continue;
             }
-
             // If the auto-start property does not have a start level,
             // then assume it is the default bundle start level, otherwise
             // parse the specified start level.
             int startLevel = sl.getInitialBundleStartLevel();
-            if (!key.equals(PROPERTY_AUTO_START)) {
+            if (!key.equals(propertyPrefix)) {
                 try {
                     startLevel = Integer.parseInt(key.substring(key.lastIndexOf('.') + 1));
                 }
@@ -427,8 +486,10 @@ public class Main {
                     System.err.println("Invalid property: " + key);
                 }
             }
-
-            StringTokenizer st = new StringTokenizer(configProps.getProperty(key), "\" ", true);
+            autoStart.put(startLevel, configProps.getProperty(key));
+        }
+        for (Integer startLevel : autoStart.keySet()) {
+            StringTokenizer st = new StringTokenizer(autoStart.get(startLevel), "\" ", true);
             if (st.countTokens() > 0) {
                 String location = null;
                 do {
@@ -438,6 +499,7 @@ public class Main {
                             String[] parts = convertToMavenUrlsIfNeeded(location, convertToMavenUrls);
                             Bundle b = context.installBundle(parts[0], new URL(parts[1]).openStream());
                             sl.setBundleStartLevel(b, startLevel);
+                            bundles.add(b);
                         }
                         catch (Exception ex) {
                             System.err.println("Auto-properties install:" + ex);
@@ -447,88 +509,7 @@ public class Main {
                 while (location != null);
             }
         }
-
-        // Now loop through and start the installed bundles.
-        for (Iterator i = configProps.keySet().iterator(); i.hasNext();) {
-            String key = (String) i.next();
-            if (key.startsWith(PROPERTY_AUTO_START)) {
-                StringTokenizer st = new StringTokenizer(configProps.getProperty(key), "\" ", true);
-                if (st.countTokens() > 0) {
-                    String location = null;
-                    do {
-                        location = nextLocation(st);
-                        if (location != null) {
-                            // Installing twice just returns the same bundle.
-                            try {
-                                String[] parts = convertToMavenUrlsIfNeeded(location, convertToMavenUrls);
-                                Bundle b = context.installBundle(parts[0], new URL(parts[1]).openStream());
-                                if (b != null) {
-                                    b.start();
-                                    checkCmProperties(b);
-                                }
-                            }
-                            catch (Exception ex) {
-                                System.err.println("Auto-properties start: " + ex);
-                            }
-                        }
-                    }
-                    while (location != null);
-                }
-            }
-        }
-    }
-
-    /**
-     * TODO: remove this hack when FELIX-1628 is properly implemented
-     * Hack for FELIX-1626.
-     * FileInstall has some delay before installing the configurations which can cause some problems
-     *
-     * @param b
-     */
-    private void checkCmProperties(final Bundle b) {
-        final String cmDir = (String) configProps.get(PROPERTY_CM_DIRECTORY);
-        if (cmProcessed || cmDir == null) {
-            return;
-        }
-        try {
-            // Try to load fileinstall internal classes.
-            // Those are not exported, so this mean if the load succeeds, we have the fileinstall bundle
-            final Class ciClass = b.loadClass("org.apache.felix.fileinstall.internal.ConfigInstaller");
-            final Class fiClass = b.loadClass("org.apache.felix.fileinstall.internal.FileInstall");
-            // If we have been able to load the classes, start a thread that will wait until
-            // fileinstall is correctly configured using config admin and push all configs.
-            new Thread() {
-                public void run() {
-                    for (int i = 0; !cmProcessed && i < 100; i++) {
-                        try {
-                            Thread.sleep(50);
-                            Method mth = fiClass.getDeclaredMethod("getConfigurationAdmin", long.class);
-                            mth.setAccessible(true);
-                            if (mth.invoke(null, 0) != null) {
-                                Constructor cns = ciClass.getDeclaredConstructor(BundleContext.class);
-                                cns.setAccessible(true);
-                                Object ci = cns.newInstance(b.getBundleContext());
-                                mth = ciClass.getDeclaredMethod("setConfig", File.class);
-                                mth.setAccessible(true);
-                                cmProcessed = true;
-//                                System.err.println("Found ready FileInstall");
-                                for (File f : new File(cmDir).listFiles()) {
-                                    if (f.getName().endsWith(".cfg")) {
-//                                        System.err.println("Processing: " + f.getName());
-                                        mth.invoke(ci, f);
-                                    }
-                                }
-                            }
-                        } catch (Throwable t) {
-                            if (cmProcessed) {
-                                t.printStackTrace();
-                            }
-                        }
-                    }
-                }
-            }.start();
-        } catch (Throwable t) {
-        }
+        return bundles;
     }
 
     private static String[] convertToMavenUrlsIfNeeded(String location, boolean convertToMavenUrls) {
@@ -634,7 +615,7 @@ public class Main {
      * arbitrary URL.
      * </p>
      */
-    private void loadSystemProperties() {
+    protected static void loadSystemProperties(File karafBase) {
         // The system properties file is either specified by a system
         // property or it is in the same directory as the Felix JAR file.
         // Try to load it from one of these places.
@@ -677,8 +658,8 @@ public class Main {
         // Perform variable substitution on specified properties.
         for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
             String name = (String) e.nextElement();
-            System.setProperty(name,
-                    substVars(props.getProperty(name), name, null, null));
+            String value = System.getProperty(name, props.getProperty(name));
+            System.setProperty(name, substVars(value, name, null, null));
         }
     }
 
@@ -714,17 +695,10 @@ public class Main {
 
         try {
             File file = new File(new File(karafBase, "etc"), CONFIG_PROPERTIES_FILE_NAME);
-            configPropURL = file.toURL();
+            configPropURL = file.toURI().toURL();
 
             file = new File(new File(karafBase, "etc"), STARTUP_PROPERTIES_FILE_NAME);
-            startupPropURL = file.toURL();
-
-            if (karafBase.equals(karafHome)) {
-                bundleDirs.add(new File(karafHome, "system"));
-            } else {
-                bundleDirs.add(new File(karafBase, "system"));
-                bundleDirs.add(new File(karafHome, "system"));
-            }
+            startupPropURL = file.toURI().toURL();
 
         }
         catch (MalformedURLException ex) {
@@ -736,6 +710,15 @@ public class Main {
         Properties configProps = loadPropertiesFile(configPropURL);
         Properties startupProps = loadPropertiesFile(startupPropURL);
 
+        String defaultRepo = System.getProperty(DEFAULT_REPO, "system");
+
+        if (karafBase.equals(karafHome)) {
+            bundleDirs.add(new File(karafHome, defaultRepo));
+        } else {
+            bundleDirs.add(new File(karafBase, defaultRepo));
+            bundleDirs.add(new File(karafHome, defaultRepo));
+        }
+
         String locations = configProps.getProperty(BUNDLE_LOCATIONS);
 
         if (locations != null) {
@@ -772,7 +755,7 @@ public class Main {
         return configProps;
     }
 
-    private static Properties loadPropertiesFile(URL configPropURL) throws Exception {
+    protected static Properties loadPropertiesFile(URL configPropURL) throws Exception {
         // Read the properties file.
         Properties configProps = new Properties();
         InputStream is = null;
@@ -801,18 +784,41 @@ public class Main {
         return configProps;
     }
 
-    private static void copySystemProperties(Properties configProps) {
+    protected static void copySystemProperties(Properties configProps) {
         for (Enumeration e = System.getProperties().propertyNames();
              e.hasMoreElements();) {
             String key = (String) e.nextElement();
             if (key.startsWith("felix.") ||
                     key.startsWith("karaf.") ||
-                    key.equals("org.osgi.framework.system.packages") ||
-                    key.equals("org.osgi.framework.bootdelegation")) {
+                    key.startsWith("org.osgi.framework.")) {
                 configProps.setProperty(key, System.getProperty(key));
             }
         }
     }
+    
+    private void updateClassLoader(Properties configProps) throws Exception {
+    	String framework = configProps.getProperty(KARAF_FRAMEWORK);
+        if (framework == null) {
+            throw new IllegalArgumentException("Property " + KARAF_FRAMEWORK + " must be set in the etc/" + CONFIG_PROPERTIES_FILE_NAME + " configuration file");
+        }
+        String bundle = configProps.getProperty(KARAF_FRAMEWORK + "." + framework);
+        if (bundle == null) {
+            throw new IllegalArgumentException("Property " + KARAF_FRAMEWORK + "." + framework + " must be set in the etc/" + CONFIG_PROPERTIES_FILE_NAME + " configuration file");
+        }
+        File bundleFile = new File(karafBase, bundle);
+        if (!bundleFile.exists()) {
+            bundleFile = new File(karafHome, bundle);
+        }
+        if (!bundleFile.exists()) {
+            throw new FileNotFoundException(bundleFile.getAbsolutePath());
+        }
+
+        URLClassLoader classLoader = (URLClassLoader) Bootstrap.class.getClassLoader();
+        Method mth = URLClassLoader.class.getDeclaredMethod("addURL", URL.class);
+        mth.setAccessible(true);
+        mth.invoke(classLoader, bundleFile.toURL());
+
+    }
 
     /**
      * Process properties to customize default felix behavior
@@ -1050,6 +1056,10 @@ public class Main {
         this.exitCode = exitCode;
     }
 
+    public Framework getFramework() {
+        return framework;
+    }
+    
     public void lock(Properties props) {
         try {
             if (Boolean.parseBoolean(props.getProperty(PROPERTY_USE_LOCK, "true"))) {
@@ -1062,6 +1072,7 @@ public class Main {
                         if (lockLogged) {
                             LOG.info("Lock acquired.");
                         }
+                        setupShutdown(props);
                         setStartLevel(defaultStartLevel);
                         for (;;) {
                             if (!lock.isAlive()) {
@@ -1101,4 +1112,121 @@ public class Main {
         sl.setStartLevel(level);
     }
 
+
+    private Random random = null;
+    private ServerSocket shutdownSocket;
+
+    protected void setupShutdown(Properties props) {
+        try {
+            String pidFile = props.getProperty(KARAF_SHUTDOWN_PID_FILE);
+            if (pidFile != null) {
+                RuntimeMXBean rtb = ManagementFactory.getRuntimeMXBean();
+                String processName = rtb.getName();
+                Pattern pattern = Pattern.compile("^([0-9]+)@.+$", Pattern.CASE_INSENSITIVE);
+                Matcher matcher = pattern.matcher(processName);
+                if (matcher.matches()) {
+                    int pid = Integer.parseInt(matcher.group(1));
+                    Writer w = new OutputStreamWriter(new FileOutputStream(pidFile));
+                    w.write(Integer.toString(pid));
+                    w.close();
+                }
+            }
+        } catch (Exception e) {
+            e.printStackTrace();
+        }
+        try {
+            int port = Integer.parseInt(props.getProperty(KARAF_SHUTDOWN_PORT, "0"));
+            String host = props.getProperty(KARAF_SHUTDOWN_HOST, "localhost");
+            String portFile = props.getProperty(KARAF_SHUTDOWN_PORT_FILE);
+            final String shutdown = props.getProperty(KARAF_SHUTDOWN_COMMAND, DEFAULT_SHUTDOWN_COMMAND);
+            if (port >= 0) {
+                shutdownSocket = new ServerSocket(port, 1, InetAddress.getByName(host));
+                if (port == 0) {
+                    port = shutdownSocket.getLocalPort();
+                }
+                if (portFile != null) {
+                    Writer w = new OutputStreamWriter(new FileOutputStream(portFile));
+                    w.write(Integer.toString(port));
+                    w.close();
+                }
+                Thread thread = new Thread() {
+                    public void run() {
+                        try {
+                            while (true) {
+                                // Wait for the next connection
+                                Socket socket = null;
+                                InputStream stream = null;
+                                try {
+                                    socket = shutdownSocket.accept();
+                                    socket.setSoTimeout(10 * 1000);  // Ten seconds
+                                    stream = socket.getInputStream();
+                                } catch (AccessControlException ace) {
+                                    LOG.log(Level.WARNING, "Karaf shutdown socket: security exception: "
+                                                       + ace.getMessage(), ace);
+                                    continue;
+                                } catch (IOException e) {
+                                    LOG.log(Level.SEVERE, "Karaf shutdown socket: accept: ", e);
+                                    System.exit(1);
+                                }
+
+                                // Read a set of characters from the socket
+                                StringBuilder command = new StringBuilder();
+                                int expected = 1024; // Cut off to avoid DoS attack
+                                while (expected < shutdown.length()) {
+                                    if (random == null) {
+                                        random = new Random();
+                                    }
+                                    expected += (random.nextInt() % 1024);
+                                }
+                                while (expected > 0) {
+                                    int ch = -1;
+                                    try {
+                                        ch = stream.read();
+                                    } catch (IOException e) {
+                                        LOG.log(Level.WARNING, "Karaf shutdown socket:  read: ", e);
+                                        ch = -1;
+                                    }
+                                    if (ch < 32) {  // Control character or EOF terminates loop
+                                        break;
+                                    }
+                                    command.append((char) ch);
+                                    expected--;
+                                }
+
+                                // Close the socket now that we are done with it
+                                try {
+                                    socket.close();
+                                } catch (IOException e) {
+                                    // Ignore
+                                }
+
+                                // Match against our command string
+                                boolean match = command.toString().equals(shutdown);
+                                if (match) {
+                                    LOG.log(Level.INFO, "Karaf shutdown socket: received shutdown command. Stopping framework...");
+                                    framework.stop();
+                                    break;
+                                } else {
+                                    LOG.log(Level.WARNING, "Karaf shutdown socket:  Invalid command '" +
+                                                       command.toString() + "' received");
+                                }
+                            }
+                        } catch (Exception e) {
+                            e.printStackTrace();
+                        } finally {
+                            try {
+                                shutdownSocket.close();
+                            } catch (IOException e) {
+                            }
+                        }
+                    }
+                };
+                thread.setDaemon(true);
+                thread.start();
+            }
+        } catch (Exception e) {
+            e.printStackTrace();
+        }
+    }
+
 }
